#include <stdio.h>
#include <time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <unistd.h>

typedef long long unsigned int uint64;

extern int trailing_zeros(int n); 
extern int trailing_zeros_asm(int n);

int print_result(int out, int out1, uint64 elapsed_time, int *);
uint64 measure_tics_per_sec();
void delay(int);

int main() {
  FILE *f = fopen("testcase", "r");
  uint64 total_time_c = 0;
  uint64 total_time_asm=0;
  uint64 ticks_per_sec = measure_tics_per_sec();
  
  int case_num = 30;
  int c_fault = 0;
  int asm_fault = 0;
  int i;
  
  printf("\n");

  // testing the first 30 numbers in "testcase"
  for (int itr = 0; itr < case_num; itr++) {
    int n;                          // input
    int out;                        // output    
    int out1;                       // output of algorithm
    unsigned int lo,hi;             // used to measure   clock cycles
    uint64 ticks, start, end;       // used for mesuring clock cycles
    
    // read one row of the testcase
    fscanf(f,"%d %d", &n, &out); 
    
    //delay(20);	   

    printf(" C : %07d", n);
    // measure clock cycles using the "rdtsc" assembly command 
    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    start = ((uint64)hi << 32) | lo;
        
    // run the algorithm
    out1 = trailing_zeros(n);

    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    end = ((uint64)hi << 32) | lo;
    ticks = end-start;
    total_time_c += ticks;
    
    print_result(out, out1, ticks, &c_fault);

    //delay(20);
    
    printf("ASM: %07d", n);
    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    start = ((uint64)hi << 32) | lo;
    
    // run the algorithm
    out1 = trailing_zeros_asm(n);
    
    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    end = ((uint64)hi << 32) | lo;
    ticks = end-start;
    
    total_time_asm += ticks;

    print_result(out, out1, ticks, &asm_fault);

    putchar('\n');

  }
  fclose(f);
  
const char msg[] = "###";
  for(int i=0; i<20; i++){
    write( STDOUT_FILENO, msg, sizeof( msg ) - 1 );
    delay(30);
  }
  putchar('\n');
  delay(500);
  
  printf("\n\t\t\x1b[32mExecuted Successfully\x1b[0m\n\n");
  printf("\t\x1b[0m");
  for(int i=0; i<37; i++)
      printf("#");
  printf("\n");
  printf("\t\x1b[0m#   \x1b[36mtotal_time ( C ) = %6f ms  \x1b[0m#\n", total_time_c / (double) (ticks_per_sec/1000));
  printf("\t\x1b[0m#   \x1b[36mtotal_time (ASM) = %.6f ms  \x1b[0m#\n", total_time_asm / (double) (ticks_per_sec/1000));
  printf("\t\x1b[0m#  \x1b[33m(C)faults:%d/%d-(ASM)faults:%d/%d\x1b[0m  #\n", c_fault, case_num, asm_fault, case_num);
  printf("\t");
  for(int i=0; i<37; i++)
      printf("#");
  printf("\x1b[0m\n\n\n");

  return 0;
}


uint64 measure_tics_per_sec() {
  unsigned int lo,hi;             // used to measure   clock cycles
  uint64 start, end;       // used for mesuring clock cycles
    
  asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    start = ((uint64)hi << 32) | lo;
        
    sleep(1);

    asm volatile ("rdtsc" : "=a" (lo), "=d" (hi));    // inline assembly
    end = ((uint64)hi << 32) | lo;

    return end-start;
}

int print_result(int out, int out1, uint64 ticks, int *fault) 
{
  printf(" -> %2d : ", out);  
  printf("elapsed= %4lld ticks ", ticks);

  int success = out == out1;
  if (success)
    printf(" \x1b[32m(PASS)\x1b[0m\n");
  else{
    printf(" \x1b[31m(FAIL!)\x1b[0m\n");
    (*fault)++;
  }

  return success;
}

void delay(int ms)
{
    // Converting time into milli_seconds
    int milli_seconds = 1000 * ms;
 
    // Stroing start time
    clock_t start_time = clock();
 
    // looping till required time is not acheived
    while (clock() < start_time + milli_seconds);
}
